## Implementation of HSSGA
## Implemented by: Arthur Guijt (arthur.guijt12@gmail.com)
## Authors: Lei He (l.he@tudelft.nl), Arthur Guijt (arthur.guijt12@gmail.com), Mathijs de Weerdt (M.M.deWeerdt@tudelft.nl), Lining Xing (xing2999@qq.com), Neil Yorke-Smith (N.Yorke-Smith@tudelft.nl)
## Date: October 20, 2019
## License: CC-BY-NC

## Implementation of HSSGA as described in:
#    Chaurasia, S. N., & Singh, A. (2017).
#    Hybrid evolutionary approaches for the single machine order acceptance and scheduling problem.
#    Applied Soft Computing, 52, 725-747.

# Original source code was not provided.
# Due to this difference performance could differ from the results listed
# in the paper.

#include("orderacceptance.jl")

OASInstance = OrderAcceptance.OASInstance

##
# Random sample function is quite useful for the mutation operator
import StatsBase:sample
# PriorityQueue allows us to subsitute the worst solution in less time
# Although the constant size probably doesn't make this matter too much.
import DataStructures:PriorityQueue, dequeue!, enqueue!

using Random


"""
    evaluateSolution(problem :: OASInstance, solution :: Vector{Int64})

Evaluate a solution

This function is a wrapper around the OrderAcceptance implementation.
"""
function evaluateSolution(problem :: OASInstance, solution :: Vector{Int64})
    evaluateSolutionX(problem, solution)[1]
end

function evaluateSolutionX(problem :: OASInstance, solution :: Vector{Int64}; selected :: Union{Nothing, BitVector}=nothing)
    n = length(solution)
    profit :: Float64 = 0.0
    Ctemp :: Int64 = typemin(Int64)
    solution′ᵢ = 0
    for i in 1:n
        @inbounds solutionᵢ = solution[i]
        # Note solution′ᵢ can be zero, make sure problem.s can be indexed for the first
        # axis with 0 by using - for example - an OffsetArray
        Ctemp′ = Ctemp
        @inbounds Ctemp = max(Ctemp, problem.r[solutionᵢ]) + problem.s[solution′ᵢ, solutionᵢ] + problem.p[solutionᵢ]
        # This is in their algorithm, but I am pretty sure this is wrong.
        # if Ctemp > problem.d̄[solution[a]]
        # Rather, the text seems to imply breaking at the first item
        # That violates the deadline.
        @inbounds if Ctemp >= problem.d̄[solutionᵢ]
            Ctemp = Ctemp′
            if selected != nothing
                selected[solutionᵢ] = false
            end
            #solution′ᵢ = solutionᵢ
            continue
        end
        @inbounds if Ctemp > problem.d[solutionᵢ]
            @inbounds profit += problem.e[solutionᵢ] + problem.w[solutionᵢ] * (problem.d[solutionᵢ] - Ctemp)
        else
            @inbounds profit += problem.e[solutionᵢ]
        end
        solution′ᵢ = solutionᵢ
    end
    # Return typemax for pastd̄i, so that we can take the minimum over blocks
    # to indicate the first index that goes past a deadline.
    return profit, Ctemp
end

"""
    AssignmentOperator(problem :: OASInstance, solution :: Vector{Int64})

Perform the assignment operator as described in the paper (section 4.7):
For all orders that have not been accepted, attempt to insert them into the schedule.
"""
function AssignmentOperator(problem :: OASInstance, solution :: Vector{Int64}; trace :: Union{Nothing, ProgressTrace})
    # Find the selected orders
    selected = in.(1:problem.n, [BitSet(solution)])
    s⃰, Clast⃰ = evaluateSolutionX(problem, solution, selected=selected)
    # Remove items with zero revenue (so that we do not end up with duplicates
    # when we insert them again...)
    filter!(f -> selected[f], solution)
    # Orders that are not selected, and should be be attempted to be inserted.
    iorders = [i for i in 1:problem.n if !selected[i]]
    # Do a shuffle so that the attempted order of insertion is random.
    shuffle!(iorders)
    # Create a working copy.
    csolution = copy(solution)
    for order in iorders
        # Reset to best solution for this order.
        # Increase size by one to accomodate new order.
        resize!(csolution, length(solution) + 1)
        copyto!(csolution, solution)
        inserted = false
        # Since the array was resized already, setting the last item rather than
        # using push!
        csolution[end] = order
        for i in length(csolution):-1:1
            s, Clast = evaluateSolutionX(problem, csolution)
            if s > s⃰ || (s == s⃰ && Clast < Clast⃰)
                s⃰ = s
                Clast⃰ = Clast
                # Update solution.
                resize!(solution, length(csolution))
                copyto!(solution, csolution)
                inserted = true
            end
            if i != 1
                # Swap.
                csolution[i-1], csolution[i] = csolution[i], csolution[i-1]
            end
        end
        # Add onto end if not inserted.
        if !inserted
            resize!(solution, length(solution)+1)
            solution[end] = order
        end
    end
    update_trace!(trace, s⃰)
    s⃰, solution
end

"""
    calculateClastandSumRevenue(problem :: OASInstance, solution :: Vector{Int64})

Calculate the the last completion time and sum of revenues.
"""
function calculateClastandSumRevenue(problem :: OASInstance, solution :: Vector{Int64})
    s, Clast = evaluateSolutionX(problem, solution)
    Clast, s
end

"""
    HSSGALS(problem :: OASInstance, solution :: Vector{Int64})

Performs a 'Local Search' as described in the paper section 4.8.
"""
function HSSGALS(problem :: OASInstance, solution :: Vector{Int64}; trace :: Union{Nothing, ProgressTrace})
     Clast⃰, s⃰ = calculateClastandSumRevenue(problem, solution)
     n = problem.n
     for i in 1:n-1
         for j in i+1:n
             solution[i], solution[j] = solution[j], solution[i]
             Clast, s = calculateClastandSumRevenue(problem, solution)
             if s > s⃰ || (s == s⃰ && Clast⃰ < Clast)
                 s⃰ = s
                 Clast⃰ = Clast
             else
                 # Swap back
                 solution[i], solution[j] = solution[j], solution[i]
             end
         end
     end
     update_trace!(trace, s⃰)
     s⃰, Clast⃰, solution
end

"""
    HSSGABTS(population :: Vector{Vector{Int64}})

Select elements from population by a binary tournament selection.
Randomly takes two elements from the population, and picks the better one with
Probability bₜ.

# Arguments
population - The population

# Parameters
bₜ          - Probability to pick the best solution in the tournament
"""
function HSSGABTS(population)
    bₜ :: Float64 = 0.50
    # Select some elements from the population.
    a, b = rand(population, 2)
    # With Probability bₜ flip the decision made (returning the worse item)
    if (a[2] > b[2]) ⊻ (rand(Float64) < bₜ)
        return a
    else
        return b
    end
end

"""
    HSSGACrossover(problem :: OASInstance, sol1 :: Vector{Int64}, sol2 :: Vector{Int64})

Perform a crossover as described in the paper.
"""
function HSSGACrossover(problem :: OASInstance, sol1 :: Vector{Int64}, sol2 :: Vector{Int64}; trace :: Union{Nothing, ProgressTrace})
    newsol = unique(ifelse.(rand(Bool, problem.n), sol1, sol2))
    AssignmentOperator(problem, newsol; trace=trace)
end

"""
    HSSGAMutation(problem :: OASInstance, sol :: Vector{Int64})

Perform a mutation as described in the paper.

# Arguments
problem - The problem the solution belongs to
sol     - The solution to mutate

# Parameters
Del - How many items to remove.
"""
function HSSGAMutation(problem :: OASInstance, sol :: Vector{Int64}; trace :: Union{Nothing, ProgressTrace})
    Del :: Int64 = min(6, problem.n)#div(problem.n, 10)+1
    @assert Del <= problem.n "You cannot remove more than the total amount of orders"
    nsol = copy(sol)
    deleteat!(nsol, sample(1:length(sol), rand(1:Del), replace=false, ordered=true))
    AssignmentOperator(problem, nsol; trace=trace)
end

"""
    HSSGAInitialSolution(problem :: OASInstance)

Generate an initial solution for the population.

# Arguments
problem - The problem instance to generate an initial solution for.
θ       - Chance it picks the item according to some heuristic number, (1-θ) chance it is picked randomly.
"""
function HSSGAInitialSolution(problem :: OASInstance, θv :: Float64; trace :: Union{Nothing, ProgressTrace})
    # Initialize empty schedule and starting time.
    sol = Vector{Int64}()
    Cₜ = 0
    O = BitSet(1:problem.n)
    k = 0
    o = nothing
    # Lazyness.
    n = problem.n
    # First order is always picked heuristically.
    θ = 1.0
    while !isempty(O)
        if rand(Float64) < θ
            # Since this changes depending on k, we cannot avoid computing this every time...
            o = argmax(ifelse.(in.(1:n, [O]), problem.e[1:n] ./ ((problem.s[k, 1:n] .+ problem.p[1:n]) .* problem.d̄[1:n]), [typemin(Float64)]))
            θ = θv
        else
            o = rand(O)
        end
        # Delete selected item from set.
        delete!(O, o)
        t = problem.s[k, o] + problem.p[o]
        # Pseudocode in paper uses Cₗ rather than Cₜ
        # As Cₗ is not defined, assuming Cₜ
        if problem.r[o] < Cₜ && Cₜ + t < problem.d̄[o]
            push!(sol, o)
            Cₜ += t
            k = o
        end
    end
    fsol, sol = AssignmentOperator(problem, sol; trace=trace)
    HSSGALS(problem, sol; trace=trace)
end

"""
    HSSGAInitialPopulation(problem :: OASInstance, size :: Int64)

Create an initial population for this instance.
"""
function HSSGAInitialPopulation(problem :: OASInstance, size :: Int64; trace :: Union{Nothing, ProgressTrace})
    θ = 0.65
    unique([HSSGAInitialSolution(problem, θ; trace=trace) for _ in 1:size])
end

function solveHSSGA(problem :: OASInstance, maxTime :: Int64; trace :: Union{Nothing, ProgressTrace} = nothing)
    population_size = 60
    πc = 0.8
    Pf = 0.05

    if maxTime < 0
        maxTime = typemax(Int64)
    end

    startTime = time()

    initialpop = HSSGAInitialPopulation(problem, population_size; trace=trace)
    # Get best solution from population.
    fs⃰, Cs⃰, s⃰ = maximum(initialpop)
    # C is negated as it does lowest first, if there are two with equal f
    # lower C is better, and thus should not be the worst.
    # By negating C this issue is resolved.
    population = PriorityQueue{Vector{Int64}, Float64}((s => f for (f, C, s) in initialpop))
    totaliters = 0
    iternotimproved = 0
    while maxTime == -1 ? (totaliters <= 10000 || iternotimproved < 5000) : ((time() - startTime)*1000 < maxTime)
        s = nothing
        totaliters += 1
        iternotimproved += 1
        if rand(Float64) <= πc
            # Perform Crossover
            sol1 = HSSGABTS(population)
            sol2 = HSSGABTS(population)
            s = HSSGACrossover(problem, sol1[1], sol2[1]; trace=trace)
        else
            # Perform mutation
            sol = HSSGABTS(population)
            s = HSSGAMutation(problem, sol[1]; trace=trace)
        end
        # Pseudocode applies AssignmentOperator again,
        # even though Crossover and Mutation already do this.

        # Destruct tuple for ease.
        fs, s = s

        if (fs⃰ - fs) < (fs * Pf)
            fs, _, s = HSSGALS(problem, s; trace=trace)
        end

        if !(s in keys(population))
            # Remove worst from population
            dequeue!(population)
            # Insert current in population
            enqueue!(population, s => fs)
            # Update best solution
            if fs > fs⃰
                fs⃰ = fs
                #Cs⃰ = Cs
                s⃰ = s
                iternotimproved = 0
            end
        end
    end
    (fs⃰, s⃰)
end
