﻿//Implementation of heuristic operators. The functions of removal operators are implemented here. The functions of insertion operators are implemented in MemeticAlgSolver.cs
//Authors: Lei He (l.he@tudelft.nl), Arthur Guijt (arthur.guijt12@gmail.com), Mathijs de Weerdt (M.M.deWeerdt@tudelft.nl), Lining Xing (xing2999@qq.com), Neil Yorke-Smith (N.Yorke-Smith@tudelft.nl)
//Date: October 20, 2019
//License: CC-BY-NC

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace MemeAlg
{
    public class Heuristic
    {
        public string type;//type of the operator: removal or insertion
        public int Heur_Id;//the id of the operator
        public double weight;//the weight of the operator
        public double score;//the score of the operator
        public int CalledTimes;//the number of called times

        /// <summary>
        /// Remove orders from current solution. The orders that are to be removed are labeled first, and then are removed in the main algorithm
        /// </summary>
        /// <param name="Id">the id of the selected removal operator</param>
        /// <param name="TargetList">the current solution</param>
        /// <param name="SizeOfBank">the number of removed orders</param>
        public void DestroyHeu(int Id, List<SelectedOrder> TargetList, int SizeOfBank)
        {
            if (SizeOfBank > TargetList.Count)
            {
                SizeOfBank = TargetList.Count - 2;
            }
            Id++;
            switch (Id)
            {
                case 1://Random removal operator
                    Remove1(TargetList, SizeOfBank);
                    break;
                case 2://Revenue removal operator
                    Remove2(TargetList, SizeOfBank);
                    break;
                case 3://Unit revenue removal operator
                    Remove3(TargetList, SizeOfBank);
                    break;
                case 4://Setup time removal operator
                    Remove4(TargetList, SizeOfBank);
                    break;
                case 5://Worst route removal operator
                    Remove5(TargetList, SizeOfBank);
                    break;
            }
        }
        

        /// <summary>
        /// Random removal operator
        /// </summary>
        /// <param name="Initial"></param>
        /// <param name="Num"></param>
        public void Remove1(List<SelectedOrder> Initial, int Num)
        {
            int n = Initial.Count;
            System.Random ran = new Random();
            int[] Random = new int[n];
            int[] index = new int[n];
            for (int j = 0; j < n; j++)
            {
                index[j] = j;
            }
            int site = n;
            int id;
            for (int j = 0; j < n; j++)
            {
                id = ran.Next(0, site - 1);
                Random[j] = index[id];
                index[id] = index[site - 1];
                site--;
            }
            id = -1;
            for (int i = 0; i < Num; i++)
            {
                id++;
                SelectedOrder obs = new SelectedOrder();
                int tag = Random[id];
                obs = Initial[tag];

                if (obs.Order.RemoveTabu >= 0 || obs.Order.IsDummy)
                    i--;
                else
                {
                    obs.ToBeRemoved = true;
                }
                if (id == n - 1)
                    break;
            }
        }

        /// <summary>
        /// Revenue removal operator
        /// </summary>
        /// <param name="Initial"></param>
        /// <param name="Num"></param>
        public void Remove2(List<SelectedOrder> Initial, int Num)
        {
            Initial = Initial.OrderBy(s => s.Order.Revenue).ToList<SelectedOrder>();
            int id = -1;
            for (int j = 0; j < Num; j++)
            {
                id++;
                SelectedOrder obs = new SelectedOrder();
                obs = Initial[id];

                if (obs.Order.RemoveTabu >= 0 || obs.Order.IsDummy)
                    j--;
                else
                {
                    obs.ToBeRemoved = true;
                }
                if (id == Initial.Count - 1)
                    break;
            }
        }

        /// <summary>
        /// Unit revenue removal operator
        /// </summary>
        /// <param name="Initial"></param>
        /// <param name="Num"></param>
        public void Remove3(List<SelectedOrder> Initial, int Num)
        {
            Initial = Initial.OrderBy(s => s.Order.UnitRevenue).ToList<SelectedOrder>();
            int id = -1;
            for (int j = 0; j < Num; j++)
            {
                id++;
                SelectedOrder obs = new SelectedOrder();
                obs = Initial[id];

                if (obs.Order.RemoveTabu >= 0 || obs.Order.IsDummy)
                    j--;
                else
                {
                    obs.ToBeRemoved = true;
                }
                if (id == Initial.Count - 1)
                    break;
            }
        }

        /// <summary>
        /// Setup time removal operator
        /// </summary>
        /// <param name="Initial"></param>
        /// <param name="Num"></param>
        public void Remove4(List<SelectedOrder> Initial, int Num)
        {
            for (int i = 1; i < Initial.Count-1;i++ )
            {
                Initial[i].Setup = Initial[i].RealStartTime - Initial[i - 1].RealEndTime;
            }

            Initial=Initial.OrderByDescending(s=>s.Setup).ToList<SelectedOrder>();
            int id = -1;
            for (int j = 0; j < Num; j++)
            {
                id++;
                SelectedOrder obs = new SelectedOrder();
                obs = Initial[id];

                if (obs.Order.RemoveTabu >= 0 || obs.Order.IsDummy)
                    j--;
                else
                {
                    obs.ToBeRemoved = true;
                }
                if (id == Initial.Count - 1)
                    break;
            }
        }

        /// <summary>
        /// Worst route removal
        /// </summary>
        /// <param name="Initial"></param>
        /// <param name="Num"></param>
        public void Remove5(List<SelectedOrder> Initial, double Num)
        {
            double Lowest = double.MaxValue;
            int flag=-1;
            double CurrentFit = 0;
            for (int i = 1; i <= Num; i++)
                CurrentFit += Initial[i].Order.Revenue;
            double UnitFit = CurrentFit / (Initial[(int)Num].RealEndTime - Initial[1].RealStartTime);

            for(int i=1;i<Initial.Count-Num;i++)
            {
                if (UnitFit < Lowest)
                {
                    Lowest = UnitFit;
                    flag = i;
                }
                CurrentFit -= Initial[i].Order.Revenue;
                CurrentFit += Initial[i +(int)Num].Order.Revenue;
                UnitFit = CurrentFit / (Initial[i + (int)Num].RealEndTime - Initial[i + 1].RealStartTime);
            }

            int id = flag-1;
            for (int j = 0; j < Num; j++)
            {
                id++;
                SelectedOrder obs = new SelectedOrder();
                obs = Initial[id];
                if (obs.Order.RemoveTabu >= 0 || obs.Order.IsDummy)
                    j--;
                else
                {
                    obs.ToBeRemoved = true;
                }
                if (id == Initial.Count - 1)
                    break;
            }
        }

        
    }
}
