## Implementation of ILS
## Implemented by: Arthur Guijt (arthur.guijt12@gmail.com)
## Authors: Lei He (l.he@tudelft.nl), Arthur Guijt (arthur.guijt12@gmail.com), Mathijs de Weerdt (M.M.deWeerdt@tudelft.nl), Lining Xing (xing2999@qq.com), Neil Yorke-Smith (N.Yorke-Smith@tudelft.nl)
## Date: October 20, 2019
## License: CC-BY-NC

## Original paper:
# Silva, Yuri Laio TV, Anand Subramanian, and Artur Alves Pessoa.
# "Exact and heuristic algorithms for order acceptance and scheduling with sequence-dependent setup times."
# Computers & Operations Research 90 (2018): 142-160.

# Original implementation is in C++ and source code was not provided.
# Due to this difference performance could differ from the results listed
# in the paper.

## Big import
include("orderacceptance.jl")
OASInstance = OrderAcceptance.OASInstance

## Imports...

# For efficiently performing a double bridge we need to sample 4 without
# duplicates, this package provides that.
import StatsBase:sample

import OffsetArrays:OffsetArray




"""
    computeCostBlock(problem :: OASInstance, b′ :: Int64, a :: Int64, b :: Int64, π :: Vector{Int64}, C_b′ :: Int64)

# Arguments
problem - The static instance data for the problem that needs to be solved.
b′      - The index in π of the last order of the previous block
a       - The index in π of the first order in the current block
b       - The index in π of the last order in the current block
π       - The vector representing the schedule of orders
C_b′    - The completion time of the last order of the previous block

# Returns
Tuple of three elements:
profit  - The profit obtained in this block
C       - Completion time of this block
pastd̄i  - The index of the first item that went past the deadline,
          MaxInt if no such item exists in a block
"""
function computeCostBlock(problem :: OASInstance,
    b′ :: Int64, a :: Int64, b :: Int64,
    π :: Vector{Int64},
    C_b′ :: Int64)
    profit :: Float64 = 0.0
    Ctemp :: Int64 = C_b′
    π′ᵢ = get(π, b′, 0)
    for i in a:b
        πᵢ = π[i]
        # Note π′ᵢ can be zero, make sure problem.s can be indexed for the first
        # axis with 0 by using - for example - an OffsetArray
        Ctemp = max(Ctemp, problem.r[πᵢ]) + problem.s[π′ᵢ, πᵢ] + problem.p[πᵢ]
        # This is in their algorithm, but I am pretty sure this is wrong.
        # if Ctemp > problem.d̄[π[a]]
        # Rather, the text seems to imply breaking at the first item
        # That violates the deadline.
        if Ctemp > problem.d̄[πᵢ]
            # Indicate which index failed.
            return profit, Ctemp, i
        end
        if Ctemp > problem.d[πᵢ]
            profit += problem.e[πᵢ] + problem.w[πᵢ] * (problem.d[πᵢ] - Ctemp)
        else
            profit += problem.e[πᵢ]
        end
        π′ᵢ = πᵢ
    end
    # Return typemax for pastd̄i, so that we can take the minimum over blocks
    # to indicate the first index that goes past a deadline.
    return profit, Ctemp, typemax(Int64)
end

# Quick alias for cost computation.
computeCostBlock(problem :: OASInstance, π :: Vector{Int64}) = computeCostBlock(problem, 0, 1, length(π), π, 0)

"""
    calculateCandG(problem :: OASInstance, π :: Vector{Int64})

Calculate the completion times and profit after every order.

# Arguments
problem - Problem that π is a solution for.
π       - A solution for `problem` that needs to be evaluated.

"""
function calculateCandG(problem :: OASInstance, π :: Vector{Int64})
    n = length(π)
    # Completion times
    C = ones(Int64, n) * typemax(Int64)
    # Profit up to a certain point
    g = zeros(Float64, n)
    profit :: Float64 = 0.0
    Ctemp :: Int64 = 0
    π′ᵢ = 0
    for i in 1:n
        πᵢ = π[i]
        # Note π′ᵢ can be zero, make sure problem.s can be indexed for the first
        # axis with 0 by using - for example - an OffsetArray
        Ctemp = max(Ctemp, problem.r[πᵢ]) + problem.s[π′ᵢ, πᵢ] + problem.p[πᵢ]
        C[i] = Ctemp
        # This is in their algorithm, but I am pretty sure this is wrong.
        # if Ctemp > problem.d̄[π[a]]
        # Rather, the text seems to imply breaking at the first item
        # That violates the deadline.
        if Ctemp > problem.d̄[πᵢ]
            # Indicate which index failed.
            C[i:n] .= Ctemp
            g[i:n] .= profit
            return C, g
        end
        if Ctemp > problem.d[πᵢ]
            profit += problem.e[πᵢ] + problem.w[πᵢ] * (problem.d[πᵢ] - Ctemp)
        else
            profit += problem.e[πᵢ]
        end
        g[i] = profit
        π′ᵢ = πᵢ
    end
    # Return typemax for pastd̄i, so that we can take the minimum over blocks
    # to indicate the first index that goes past a deadline.
    return C, g
end

"""
    computeCostSwap(problem :: OASInstance, π :: Vector{Int64}, i :: Int64, j :: Int64,
        g :: Vector{Float64}, C :: Vector{Int64})

# Important
Preferably ensure that i < j.

# Arguments
problem - The instance which is being solved.
π       - The solution on which a swap needs to be evaluated.
i       - The first index of the items to be swapped.
j       - The second index of the items to be swapped.
g       - A vector of the 'costs' for π for up to each index.
C       - A vector of the completion times for π up to each index.

The latter two are used to avoid unneccesary recomputation as up to the point
of the swap, the evaluation is the same.

# Returns
A single value
f - The total revenue after the swap
"""
function computeCostSwap(problem :: OASInstance, π :: Vector{Int64},
        i :: Int64, j :: Int64,
        g :: Vector{Float64}, C :: Vector{Int64})
    # To ensure i < j as described in the paper.
    i = min(i, j)
    j = max(i, j)
    n = length(π)
    # Skip computation of first block by using f and Ctemp.
    f = get(g, i-1, 0)
    Ctemp = get(C, i-1, 0)
    # Swapped item j
    fblock, Ctemp, pastd̄i = computeCostBlock(problem, i-1, j, j, π, Ctemp)
    f += fblock
    if pastd̄i != typemax(Int64)
        return f
    end
    # Block from i+1 to j-1
    fblock, Ctemp, pastd̄i = computeCostBlock(problem, j, i+1, j-1, π, Ctemp)
    f += fblock
    if pastd̄i != typemax(Int64)
        return f
    end
    # Swapped item i
    # Error in paper: Note that precursor at j-1 should be j if j-1 == i
    # This case is not handled properly in the paper.
    fblock, Ctemp, pastd̄i = computeCostBlock(problem, ifelse(j-1 == i, j, j-1), i, i, π, Ctemp)
    f += fblock
    if pastd̄i != typemax(Int64)
        return f
    end
    # Block from j+1 till end
    fblock, Ctemp, pastd̄i = computeCostBlock(problem, i, j+1, n, π, Ctemp)
    f += fblock
    return f
end

"""
    neighbourhoodSwap(problem :: OASInstance, π :: Vector{Int64},
        g :: Vector{Float64}, C :: Vector{Int64},
        Δs_swap :: Vector{Int64}, maxΔs_swap :: Int64)

# Arguments
problem    - The instance which is being solved.
π          - The solution on which a swap needs to be evaluated.
g          - A vector of the 'costs' for π for up to each index.
C          - A vector of the completion times for π up to each index.
Δs_swap    - Vector used for learning purposes for later maxΔs_swap values.
             When not learning: probably empty!
maxΔs_swap - The maximum setup variation allowed before evaluation.
             Avoids evaluating probably bad solutions.
"""
function neighbourhoodSwap(problem :: OASInstance, π :: Vector{Int64},
        g :: Vector{Float64}, C :: Vector{Int64},
        Δs_swap :: Vector{Int64}, maxΔs_swap :: Int64)
    n = length(π)
    # Can possibly be replaced by g[end] and an additional argument.
    f⃰, _, pastd̄i = computeCostBlock(problem, 0, 1, n, π, 0)
    fo = f⃰
    N = Vector{Tuple{Float64, Vector{Int64}}}()
    s = problem.s
    for i in 1:min(pastd̄i-1, n)
        for j in i+1:min(pastd̄i-1, n)
            setupVariation =
                +(s[i-1, j], s[j, i+1], s[j-1, i], s[i, j+1]) -
                +(s[i-1, i], s[i, i+1], s[j-1, j], s[j, j+1])
            if setupVariation < maxΔs_swap
                fπ′ = computeCostSwap(problem, π, i, j, g, C)
                if fπ′ > fo
                    # Solution improves over original, add to neighbourhood.
                    # Immidiately create the solution
                    # Materializing the actual solution later (when selected)
                    # is probably faster though!
                    π′ = copy(π)
                    π′[[i, j]] .= π[[j, i]]

                    # Add solution to neighbourhood.
                    push!(N, (fπ′, π′))
                    # Collect data when 'training'
                    if maxΔs_swap == typemax(Int64)
                        push!(Δs_swap, setupVariation)
                    end
                end
            end
        end
    end
    return N
end

"""
    computeCostlBlockForward(problem :: OASInstance, π :: Vector{Int64},
        i :: Int64, j :: Int64, l :: Int64,
        g :: Vector{Float64}, C :: Vector{Int64})

# Important
This method assumes i < j (as otherwise the relocation is not forwards)

# Arguments
problem - The instance which is being solved.
π       - The solution on which a swap needs to be evaluated.
i       - The position of the block
j       - The new position of the block
l       - The length of the block
g       - A vector of the 'costs' for π for up to each index.
C       - A vector of the completion times for π up to each index.
"""
function computeCostlBlockForward(problem :: OASInstance, π :: Vector{Int64},
        i :: Int64, j :: Int64, l :: Int64,
        g :: Vector{Float64}, C :: Vector{Int64})
    # Give name for convinience.
    n = length(π)
    # Skip computation of first block by using f and Ctemp.
    f = get(g, i-1, 0)
    Ctemp = get(C, i-1, 0)
    # Skip items in block and continue with next block until insertion point.
    fblock, Ctemp, pastd̄i = computeCostBlock(problem, i-1, i+l, j, π, Ctemp)
    f += fblock
    if pastd̄i != typemax(Int64)
        return f
    end
    # Insert the block.
    fblock, Ctemp, pastd̄i = computeCostBlock(problem, j, i, i-1+l, π, Ctemp)
    f += fblock
    if pastd̄i != typemax(Int64)
        return f
    end
    #pastd̄i = min(pastd̄i, pastd̄iblock)
    # Add remainder
    fblock, Ctemp, pastd̄i = computeCostBlock(problem, i-1+l, j+1, n, π, Ctemp)
    f += fblock
    #pastd̄i = min(pastd̄i, pastd̄iblock)
    return f#, pastd̄iblock
end

"""
    computeCostlBlockBackward(problem :: OASInstance, π :: Vector{Int64},
        i :: Int64, j :: Int64, l :: Int64,
        g :: Vector{Float64}, C :: Vector{Int64})

# Important
This method assumes i > j (as otherwise the relocation is not backwards)

# Arguments
problem - The instance which is being solved.
π       - The solution on which a swap needs to be evaluated.
i       - The position of the block
j       - The new position of the block
l       - The length of the block
g       - A vector of the 'costs' for π for up to each index.
C       - A vector of the completion times for π up to each index.
"""
function computeCostlBlockBackward(problem :: OASInstance, π :: Vector{Int64},
        i :: Int64, j :: Int64, l :: Int64,
        g :: Vector{Float64}, C :: Vector{Int64})
    # Give name for convinience.
    n = length(π)
    # Skip computation of first block by using f and Ctemp.
    f = get(g, j-1, 0)
    Ctemp = get(C, j-1, 0)
    # Insert block.
    fblock, Ctemp, pastd̄i = computeCostBlock(problem, j-1, i, i-1+l, π, Ctemp)
    f += fblock
    if pastd̄i != typemax(Int64)
        return f
    end
    # Add block from insertion point to block location.
    fblock, Ctemp, pastd̄i = computeCostBlock(problem, i-1+l, j, i-1, π, Ctemp)
    f += fblock
    if pastd̄i != typemax(Int64)
        return f
    end
    #pastd̄i = min(pastd̄i, pastd̄iblock)
    # Add the block from after moved block till end.
    fblock, Ctemp, pastd̄i = computeCostBlock(problem, i-1, i+l, n, π, Ctemp)
    f += fblock
    if pastd̄i != typemax(Int64)
        return f
    end
    #pastd̄i = min(pastd̄i, pastd̄iblock)
    return f#, pastd̄iblock
end

"""
    neighbourhoodLBlockInsert(problem :: OASInstance, π :: Vector{Int64},
        l :: Int64,
        g :: Vector{Float64}, C :: Vector{Int64},
        Δs_insert :: Vector{Int64}, maxΔs_insert :: Int64)

# Arguments
problem      - The instance which is being solved.
π            - The solution on which a swap needs to be evaluated.
l            - The length of the block insertion area
g            - A vector of the 'costs' for π for up to each index.
C            - A vector of the completion times for π up to each index.
Δs_insert    - Vector used for learning purposes for later maxΔs_insert values.
               When not learning: probably empty!
maxΔs_insert - The maximum setup variation allowed before evaluation.
               Avoids evaluating probably bad solutions.
"""
function neighbourhoodLBlockInsert(problem :: OASInstance, π :: Vector{Int64},
        l :: Int64,
        g :: Vector{Float64}, C :: Vector{Int64},
        Δs_insert :: Vector{Int64}, maxΔs_insert :: Int64)
    n = length(π)
    # Can possibly be replaced by g[end] and an additional argument.
    f⃰, _, pastd̄i = computeCostBlock(problem, 0, 1, n, π, 0)
    fo = f⃰
    N = Vector{Tuple{Float64, Vector{Int64}}}()
    s = problem.s
    # Forwards
    for i in 1:min(n, pastd̄i)-l
        # Paper says j in i+1:n, though this yields some issues...
        for j in i+l:n
            setupVariation =
                +(s[i-1, i+l], s[j, i], s[i-1+l, j+1]) -
                +(s[i-1, i], s[i-1+l, i+l], s[j, j+1])
            if setupVariation <= maxΔs_insert
                fπ′ = computeCostlBlockForward(problem, π, i, j, l, g, C)
                if fπ′ > fo
                    # Materialize solution.
                    π′ = copy(π)
                    π′[i:j-l] .= π[i+l:j]
                    π′[j-l+1:j] .= π[i:i+l-1]
                    push!(N, (fπ′, π′))
                    # Collect info for 'learning'
                    if maxΔs_insert == typemax(Int64)
                        push!(Δs_insert, setupVariation)
                    end
                end
            end
        end
    end
    # Backwards
    for i in 1+l:n-l+1
        for j in 1:min(i, pastd̄i)-1
            setupVariation =
                +(s[j-1, i], s[j-1+l, j], s[i-1, i+1]) -
                +(s[j-1,j], s[i-1, i], s[i-1+l, i+1])
            if setupVariation <= maxΔs_insert
                fπ′ = computeCostlBlockBackward(problem, π, i, j, l, g, C)
                if fπ′ > fo
                    # Materialize solution.
                    π′ = copy(π)
                    π′[j:j+l-1] .= π[i:i-1+l]
                    π′[j+l:i+l-1] .= π[j:i-1]

                    push!(N, (fπ′, π′))
                    # Collect info for 'learning'
                    if maxΔs_insert == typemax(Int64)
                        push!(Δs_insert, setupVariation)
                    end
                end
            end
        end
    end
    N
end

"""
    setupVariationToMaxDelta(Δs :: Vector{Int64}, θ :: Float64)

# Arguments
Δs - The vector of setupVariations.
θ  - Number in [0, 1] where a lower value drops more of the neighbourhood.
"""
function setupVariationToMaxDelta(Δs :: Vector{Int64}, θ :: Float64)
    # Sort ascending, as θ higher selects higher indices.
    # This index should allow more to pass through than for lower values of θ.
    sort!(Δs)
    if length(Δs) == 0
        return typemax(Int64)
    end
    Δs[floor(Int64, length(Δs) * θ)]
end

"""
    apply(problem :: OASInstance, solution :: Vector{Int64})

Unlike normal Local Search, RNVD takes the neighbourhood of improving moves and
rather than picking the best improvement, picks a *random* improvment.
"""
function applyRNVD(problem :: OASInstance, solution :: Vector{Int64},
        Δs_𝒩 :: Vector{Vector{Int64}}, maxΔs_𝒩 :: Vector{Int64})
    π = solution
    prevg = -9999.0
    ix = 0
    while true
        C, g = calculateCandG(problem, π)

        if prevg >= g[end]
            #error("Problem: $(prevg) >= $(g[end]), LS made worsening step.")
            ix += 1
        end
        if ix > 15
            return π, g[end]
        end
        #fπ = g[end]
        N = vcat(
            # Swap neighbourhood
            neighbourhoodSwap(problem, π, g, C, Δs_𝒩[1], maxΔs_𝒩[1]),
            # Block neighbourhoods for lengths 1, 2 and 3.
            neighbourhoodLBlockInsert(problem, π, 1, g, C, Δs_𝒩[2], maxΔs_𝒩[2]),
            neighbourhoodLBlockInsert(problem, π, 2, g, C, Δs_𝒩[3], maxΔs_𝒩[3]),
            neighbourhoodLBlockInsert(problem, π, 3, g, C, Δs_𝒩[4], maxΔs_𝒩[4])
        )
        if length(N) == 0
            return π, g[end]
        end
        fπ, π = rand(N)
        prevg = g[end]
    end
end

"""
    generateRandomSolution(problem :: OASInstance)

Generate a random solution for a problem.
"""
function generateRandomSolutionSilva(problem :: OASInstance)
    sortperm(rand(Float64, problem.n))
end


function generateInitialSolutionSilva(problem :: OASInstance)
    generateRandomSolutionSilva(problem)
end

"""
    applyPerturbation(π :: Vector{Int64})

Apply a double bridge reminicent of the double bridge neighbourhood used for
TSP problems.

# Arguments
π - The solution that needs to be perturbed.
"""
function applyPerturbation(π :: Vector{Int64})
    edges = sample(2:length(π), 4, replace=false, ordered=true)
    # Perform the actual double bridge.
    π′ = similar(π)
    p = edges[1]-1
    π′[1:p] .= π[1:edges[1]-1]
    p′ = p+1
    p += edges[4]-edges[3]
    π′[p′:p] .= π[edges[3]:edges[4]-1]
    p′ = p+1
    p += edges[3]-edges[2]
    π′[p′:p] .= π[edges[2]:edges[3]-1]
    p′ = p+1
    p += edges[2]-edges[1]
    π′[p′:p] .= π[edges[1]:edges[2]-1]
    p += 1
    π′[p:end] .= π[edges[4]:end]

    return π′
end
"""
    solveILSOAS(problem :: OASInstance, I_R :: Int64, I_ILS :: Int64)

V
"""
function solveILSOAS(problem :: OASInstance, I_R :: Int64, I_ILS :: Int64)
    # General structure
    # Perform the following I_R times.
    # 1. Generate an initial feasible solution - Currently random.
    # 2. Improve it using the RNVD procedure - ✓
    # 3. Perform the following procedure until there was no improvement I_ILS times - ✓
    #   a. Apply perturbation - ✓
    #   b. Perform RNVD procedure on solution - ✓
    #   c. Don't forget to update best solution if it improves. - ✓
    # 4. If the current solution is better than the current overal solution
    #     update this overal solution. - ✓
    θ = 0.8
    # Amount of neighbourhoods
    𝒩 = 4
    maxΔs_𝒩 = [typemax(Int64) for i in 1:4]
    Δs_𝒩 = [Vector{Int64}() for i in 1:4]
    sol⃰ = collect(1:problem.n)
    fsol⃰ = 0
    ST=time()
    for iter in 1:I_R
        sol = generateInitialSolutionSilva(problem)
        sol, fsol = applyRNVD(problem, sol, Δs_𝒩, maxΔs_𝒩)
        sol′, fsol′ = sol, fsol
        iter_ILS = 0

        while iter_ILS < I_ILS
            sol′ = applyPerturbation(sol′)
            sol′, fsol′n = applyRNVD(problem, sol′, Δs_𝒩, maxΔs_𝒩)
            iter_ILS += 1
            CT=time()
            if CT-ST>3600
                break
            end
            if fsol < fsol′
                iter_ILS = 0
                sol = sol′
                fsol = fsol′
            end
        end
        # Update global solution if the best result of this iteration resulted
        # in something better.
        if fsol > fsol⃰
            fsol⃰ = fsol
            sol⃰ = sol
        end
        CT=time()
        if CT-ST>3600
            break
        end
        # The first round of Iterated Local Search is also used for
        # finding a good maxΔs_𝒩.
        if iter == 1
            for 𝒩i in 1:𝒩
                maxΔs_𝒩[𝒩i] = setupVariationToMaxDelta(Δs_𝒩[𝒩i], θ)
            end
            # To save memory.
            Δs_𝒩 = [Vector{Int64}() for i in 1:4]
        end
    end
    fsol⃰, sol⃰
end

## Debugging scripty.
problem = OrderAcceptance.parseInstance("./Dataslack_10orders_Tao1R1_1.txt", 10)
randomSolution = generateRandomSolutionSilva(problem)

p1, C1, d1 = computeCostBlock(problem, 0, 1, problem.n, randomSolution, 0)

function test()
    p2 = 0
    cn  = 0
    for i in 1:problem.n
        pb, cn, db = computeCostBlock(problem, i-1, i, i, randomSolution, cn)
        cx2 = cn
        p2 += pb
        if db != typemax(Int64)
            return p2, cn, p2 - pb
        end
    end
    return p2, cn
end
# test()
