﻿//The memetic algorithm
//Authors: Lei He (l.he@tudelft.nl), Arthur Guijt (arthur.guijt12@gmail.com), Mathijs de Weerdt (M.M.deWeerdt@tudelft.nl), Lining Xing (xing2999@qq.com), Neil Yorke-Smith (N.Yorke-Smith@tudelft.nl)
//Date: October 20, 2019
//License: CC-BY-NC

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Diagnostics;
using System.IO;

namespace MemeAlg
{
    class MemeticAlgSolver
    {
        public problem Problem = new problem();

        //Parameters
        public Random r = new Random(System.DateTime.Now.Millisecond);
        public int SizeofPop = 20;
        public int MaxIte = 50000;
        public int jumpout = 10;

        public double OffSpringRatio = 0.4;
        public double MutationRatio = 0.1;
        public double EliteRatio = 0.5;
        
        int SigmaGlobal = 30; //The score an operator wins when it finds a global best solution
        int SigmaLocal = 20; //The score an operator wins when it finds a better solution
        int SigmaBad = 10; //The score an operator wins when it finds a slightly worse solution and the solution is accepted
        int tabutime; //When an ordre is inserted in the current solution, its removal is forbidden for the following 'tabutime' iterations
        double per = 0.1; //The percentage of orders removed by ALNS in the beginning
        double Lam = 0.5; //The weight update parameter
        double Temperature = 0;
        double ParaOfAnnealing = 0.9975;
        double GoodPiece = 2; //The parameter of good small piece. A small piece consists of two adjacent orders. And if the unit revenue (total revenue divided by total processing time and the gap) is above 2, it is recognized as a good piece.
        public List<double> GeneValueLib = new List<double>();
        public List<Heuristic> DestroyList = new List<Heuristic>(); //All the removal operators
        public List<Heuristic> InsertList = new List<Heuristic>(); //All the repair operators
        double firstgoodpiece = 0;
        //Parameters

        /// <summary>
        /// The main function of the algorithm
        /// </summary>
        /// <param name="time">Anytime time value</param>
        /// <param name="rev">Anytime revenue</param>
        public void StartExecute(out List<double> time, out List<double> rev)
        {
            List<Chromosome> InitialPop = new List<Chromosome>();
            List<Chromosome> Elite = new List<Chromosome>();
            List<Chromosome> NonElite = new List<Chromosome>();
            time = new List<double>();
            rev = new List<double>();

            //preprocess
            double TotalLength = Problem.OrderList[Problem.OrderList.Count - 1].Deadline;
            foreach (Order a in Problem.OrderList)
            {
                a.low = a.Release * 1.0 / TotalLength;
                a.high = a.Deadline * 1.0 / TotalLength;
            } //The random key of each order is generated between a.low and a.high

            Problem.SelectedOrderList = CreateSelectedOrderList(Problem.OrderList);
            InitialOperatorPool();
            tabutime = (int)Math.Sqrt(Problem.OrderList.Count / 2);
            int noimprove = 0;
            double G_fit = 0;

            double step = 0.99 / (Problem.OrderList.Count - 2);
            for (int i = 0; i < Problem.OrderList.Count - 1; i++)
                GeneValueLib.Add(i * step);
            GeneValueLib.Add(1);
            int crossbest = 0;
            //preprocess

            Stopwatch st = new Stopwatch();
            st.Start();

            InitialPop = InitializePopulation();

            //Enter the iteration
            for (int it = 0; it < MaxIte; it++)
            {
                noimprove++;

                for (int i = 0; i < InitialPop.Count; i++)
                {
                    //Decode first. the Fitness and the solution is stored in the chromosome
                    if (r.NextDouble() > Problem.Instance.AvgWindowLength*2)
                        ComplexDecode(InitialPop[i]);
                    else
                        Decode(InitialPop[i]);

                    //initialize the temperature
                    if (i == 0 && it == 0)
                    {
                        Temperature = -(0.05 / Math.Log(0.5) * InitialPop[i].Fitness);
                        if (Temperature > 100)
                            Temperature = 100;
                    }
                    if (InitialPop[i].Fitness > G_fit * 0.9)
                        ALNSMain(InitialPop[i], it, G_fit, false); //Use ALNS to improve the popululation (After the first iteration, the population here only contains the new generation (only the crossover set))
                    if (InitialPop[i].Fitness > G_fit)
                    {
                        if (it != 0)
                            crossbest++;
                        G_fit = InitialPop[i].Fitness;
                        noimprove = 0;
                    }
                }

                for (int i = 0; i < Elite.Count; i++)
                {
                    ALNSMain(Elite[i], it, G_fit, true); //Use ALNS to improve the Elite set
                    if (Elite[i].Fitness > G_fit)
                    {
                        G_fit = Elite[i].Fitness;
                        noimprove = 0;
                    }
                }

                AddListToList(InitialPop, Elite);

                InitialPop = InitialPop.OrderByDescending(s => s.Fitness).ToList<Chromosome>();

                Elite.Clear();
                NonElite.Clear();

                for (int i = 0; i < SizeofPop; i++)
                {
                    if (i < SizeofPop * EliteRatio)
                        Elite.Add(InitialPop[i]);
                    else
                        NonElite.Add(InitialPop[i]);
                }

                List<Chromosome> CrossOverSet = new List<Chromosome>();
                CrossOverSet = RepairCrossOver(Elite, NonElite);
                List<Chromosome> MutantSet = Mutation();

                InitialPop = CrossOverSet;
                AddListToList(InitialPop, MutantSet);

                //Update the weights of different operators of ALNS
                if (it % 6 == 5)
                {
                    UpdateHeuristicList(DestroyList, Lam);
                    UpdateHeuristicList(InsertList, Lam);
                }

                rev.Add(G_fit);
                time.Add(st.Elapsed.TotalSeconds);

                if (noimprove > jumpout * (Problem.OrderList.Count-2)|| G_fit == Problem.TotalRev)
                    break;
            }
            st.Stop();

            Problem.CheckSolution(Elite[0].Solution);

            while (time.Count <= MaxIte)
            {
                time.Add(st.Elapsed.TotalSeconds);
                rev.Add(G_fit);
            }
        }

        
        
        /// <summary>
        /// Initialization randomly
        /// </summary>
        /// <returns></returns>
        List<Chromosome> InitializePopulation()
        {
            List<Chromosome> InitialPop = new List<Chromosome>();

            for (int i = 0; i < SizeofPop; i++)
            {
                Chromosome NewChromo = new Chromosome();
                for (int j = 0; j < Problem.OrderList.Count; j++)
                {
                    Gene NewGene = new Gene();
                    Order Cur = Problem.OrderList[j];
                    NewGene.OrderId = j;
                    NewGene.GeneValue = Cur.high - (Cur.high - Cur.low) * r.NextDouble();
                    NewChromo.GeneList.Add(NewGene);
                }
                InitialPop.Add(NewChromo);
            }
            return InitialPop;
        }

        /// <summary>
        /// Mutation process
        /// </summary>
        /// <returns></returns>
        List<Chromosome> Mutation()
        {
            //randomly/greedily
            List<Chromosome> InitialPop = new List<Chromosome>();

            for (int i = 0; i < SizeofPop * MutationRatio; i++)
            {
                Chromosome NewChromo = new Chromosome();
                for (int j = 0; j < Problem.OrderList.Count; j++)
                {
                    Gene NewGene = new Gene();
                    Order Cur = Problem.OrderList[j];
                    NewGene.OrderId = j;
                    NewGene.GeneValue = (Cur.high - (Cur.high - Cur.low) * r.NextDouble());
                    NewChromo.GeneList.Add(NewGene);
                }
                InitialPop.Add(NewChromo);
            }
            return InitialPop;
        }

        /// <summary>
        /// Decode from the lowest random key.
        /// </summary>
        /// <param name="OneChromo"></param>
        void Decode(Chromosome OneChromo)
        {
            List<Gene> tempGeneList = OneChromo.GeneList.OrderBy(s => s.GeneValue).ToList<Gene>();
            List<SelectedOrder> Solution = new List<SelectedOrder>();
            OneChromo.UnScheduledOrders.Clear();

            for (int i = 0; i < tempGeneList.Count; i++)
            {
                SelectedOrder newselorder = new SelectedOrder();
                newselorder.Order = Problem.OrderList[tempGeneList[i].OrderId];

                if (i == 0)
                    newselorder.RealStartTime = 0;
                else if (i == tempGeneList.Count - 1)
                    newselorder.RealStartTime = newselorder.Order.Release;
                else
                    newselorder.RealStartTime = Problem.GetEarliestTime(Solution[Solution.Count - 1], newselorder);

                if (newselorder.RealStartTime > newselorder.Order.Deadline)
                {
                    OneChromo.UnScheduledOrders.Add(newselorder);
                    continue;
                }

                newselorder.RealEndTime = newselorder.RealStartTime + newselorder.Order.ProcessTime;
                Solution.Add(newselorder);
            }
            OneChromo.Solution = Solution;
            //Problem.CheckSolution(Solution);
            OneChromo.Fitness = Problem.GetTotalProfit(Solution);
        }

        void ComplexDecode(Chromosome OneChromo)
        {
            List<Gene> tempGeneList = OneChromo.GeneList.OrderBy(s => s.GeneValue).ToList<Gene>();
            List<SelectedOrder> Solution = new List<SelectedOrder>();
            OneChromo.UnScheduledOrders.Clear();
            SelectedOrder startorder = new SelectedOrder();
            startorder.Order = Problem.OrderList[0];
            startorder.RealStartTime = 0;
            startorder.RealEndTime = 0;
            Solution.Add(startorder);
            SelectedOrder endorder = new SelectedOrder();
            endorder.Order = Problem.OrderList[Problem.OrderList.Count - 1];
            endorder.RealStartTime = endorder.Order.Release;
            endorder.RealEndTime = endorder.Order.Release;
            Solution.Add(endorder);

            double TimePoint;
            double TimePointEnd = 0;
            double folTimePoint = 0;
            List<SelectedOrder> BankList = new List<SelectedOrder>();

            for (int i = 1; i < tempGeneList.Count - 1; i++)
            {
                SelectedOrder newselorder = new SelectedOrder();
                newselorder.Order = Problem.OrderList[tempGeneList[i].OrderId];
                bool insert = false;

                for (int j = 1; j < Solution.Count; j++)
                {
                    if (Solution[j].Order.Deadline > newselorder.Order.Release)
                    {
                        TimePoint = Problem.GetEarliestTime(Solution[j - 1], newselorder);
                        TimePointEnd = TimePoint + newselorder.Order.ProcessTime;
                        folTimePoint = Problem.GetTempEarliestTime(TimePointEnd, newselorder, Solution[j]);

                        if (TimePoint > newselorder.Order.Deadline || folTimePoint - Solution[j].RealStartTime > Solution[j].TS)//Time slack is not enough
                            continue;

                        Solution.Insert(j, newselorder);
                        insert = true;
                        UpdateSolution(ref Solution, BankList);
                        break;
                    }
                }

                if (!insert)
                {
                    OneChromo.UnScheduledOrders.Add(newselorder);
                    continue;
                }
            }
            OneChromo.Solution = Solution;
            OneChromo.Fitness = Problem.GetTotalProfit(Solution);
        }

        List<Chromosome> RepairCrossOver(List<Chromosome> Elite, List<Chromosome> NonElite)
        {
            List<Chromosome> OffSpring = new List<Chromosome>();

            for (int i = 0; i < OffSpringRatio * SizeofPop; i++)
            {
                Chromosome Dad = NonElite[r.Next(NonElite.Count)];
                Chromosome Mom = Elite[r.Next(Elite.Count)];
                Chromosome Son = new Chromosome();

                for (int j = 0; j < Problem.OrderList.Count; j++)
                {
                    Gene newgene = new Gene();
                    newgene.OrderId = j;
                    if (!Mom.GeneList[j].IsPairChosen && !Dad.GeneList[j].IsPairChosen)
                    {
                        if (r.NextDouble() <= 0.7)
                        {
                            newgene.GeneValue = Mom.GeneList[j].GeneValue;
                            if (Mom.GeneList[j].IsPairedPre)
                                Mom.GeneList[j].PairGenePre.IsPairChosen = true;
                            if (Mom.GeneList[j].IsPairedSuc)
                                Mom.GeneList[j].PairGeneSuc.IsPairChosen = true;
                        }
                        else
                        {
                            newgene.GeneValue = Dad.GeneList[j].GeneValue;
                            if (Dad.GeneList[j].IsPairedPre)
                                Dad.GeneList[j].PairGenePre.IsPairChosen = true;
                            if (Dad.GeneList[j].IsPairedSuc)
                                Dad.GeneList[j].PairGeneSuc.IsPairChosen = true;
                        }
                    }
                    else if (Mom.GeneList[j].IsPairChosen)
                    {
                        if (r.NextDouble() < 0.95)
                            newgene.GeneValue = Mom.GeneList[j].GeneValue;
                        else
                            newgene.GeneValue = Dad.GeneList[j].GeneValue;
                    }
                    else if (Dad.GeneList[j].IsPairChosen)
                    {
                        if (r.NextDouble() < 0.95)
                            newgene.GeneValue = Dad.GeneList[j].GeneValue;
                        else
                            newgene.GeneValue = Mom.GeneList[j].GeneValue;
                    }
                    Son.GeneList.Add(newgene);
                }
                OffSpring.Add(Son);
            }

            return OffSpring;
        }
        
        /// <summary>
        /// Add list B to the end of list A
        /// </summary>
        /// <param name="A"></param>
        /// <param name="B"></param>
        void AddListToList(List<Chromosome> A, List<Chromosome> B)
        {
            for (int i = 0; i < B.Count; i++)
                A.Add(B[i]);
        }

        /// <summary>
        /// change class order to class selected order
        /// </summary>
        /// <param name="ol"></param>
        /// <returns></returns>
        public List<SelectedOrder> CreateSelectedOrderList(List<Order> ol)
        {
            List<SelectedOrder> List = new List<SelectedOrder>();

            for (int i = 0; i < ol.Count; i++)
            {
                SelectedOrder Cur_obs = new SelectedOrder();
                Cur_obs.RealStartTime = ol[i].Release;
                Cur_obs.RealEndTime = Cur_obs.RealStartTime + ol[i].ProcessTime;
                Cur_obs.Order = ol[i];
                List.Add(Cur_obs);
            }
            return List;
        }


        /// <summary>
        /// ALNS search process
        /// </summary>
        /// <param name="IniSol">Initial solution</param>
        /// <param name="CurrentIte">Current iteration time</param>
        /// <param name="G_fit">the fitness</param>
        /// <param name="IsElite">whether the individual is elite</param>
        public void ALNSMain(Chromosome IniSol, int CurrentIte, double G_fit, bool IsElite)
        {
            //Update Tabu attribute of each order
            for (int ii = 0; ii < Problem.OrderList.Count; ii++)
            {
                Problem.OrderList[ii].RemoveTabu--;
            }

            List<Gene> CurrentGeneList = new List<Gene>();
            List<SelectedOrder> NewCurrent = new List<SelectedOrder>();//The new solution
            List<SelectedOrder> BankList = new List<SelectedOrder>();//The candidate tasks set
            List<SelectedOrder> UnScheduledOrders = new List<SelectedOrder>(); //For storing all failed tasks
            CopyListToList(IniSol.UnScheduledOrders, BankList);//Current unscheduled tasks plus the removed tasks, will become the new candidate task set

            if (!IsElite)
                CurrentIte = 0;

            int SizeOfRequestBank = (int)Math.Ceiling(IniSol.Solution.Count * per * (MaxIte - CurrentIte) * 1.0 / MaxIte);

            if (SizeOfRequestBank > IniSol.Solution.Count - 2)
                SizeOfRequestBank = IniSol.Solution.Count - 2;

            double C_fit = IniSol.Fitness; //Current fitness
            if (G_fit == 0)
                G_fit = C_fit;
            double NC_fit = 0;//The new fitness

            Heuristic CurDelete = new Heuristic();
            CurDelete.type = "Destroy";
            Heuristic CurInsert = new Heuristic();
            CurInsert.type = "Repair";
            int SelectDelNum, SelectInsNum;

            CloneGeneList(IniSol.GeneList, CurrentGeneList);//backup the genelist

            SelectDelNum = SelectOfHeuristic(DestroyList);//select a removal operator
            CurDelete = DestroyList[SelectDelNum];
            CurDelete.CalledTimes++;

            CopyListToList(IniSol.Solution, NewCurrent);
            CurDelete.DestroyHeu(CurDelete.Heur_Id, NewCurrent, SizeOfRequestBank);//select the orders to be removed

            List<Order> JustRemove = new List<Order>();
            CheckRemove(IniSol.Solution, NewCurrent, BankList, CurrentGeneList, JustRemove); //remove the orders and add them to the banklist

            if (NewCurrent.Count > 0)
                UpdateSolution(ref NewCurrent, BankList);//update the solution and time slacks          

            double Upperbound = Problem.GetTotalProfit(NewCurrent);
            double BankRevenue = 0;
            foreach (SelectedOrder a in BankList)
                BankRevenue += a.Order.Revenue;


            //**************????***********************
            SelectInsNum = SelectOfHeuristic(InsertList);
            CurInsert = InsertList[SelectInsNum];//select an insertion operator
            CurInsert.CalledTimes++;


            switch (CurInsert.Heur_Id)
            {
                case 0:

                    for (int inser = 0; inser < BankList.Count; inser++)
                        BankList[inser].HeuristicInfo = BankList[inser].Order.Revenue * (1 + r.NextDouble());
                    BankList = BankList.OrderByDescending(s => s.HeuristicInfo).ToList<SelectedOrder>();
                    break;
                case 1:
                    for (int inser = 0; inser < BankList.Count; inser++)
                        BankList[inser].HeuristicInfo = BankList[inser].Order.UnitRevenue * (1 + r.NextDouble());
                    BankList = BankList.OrderByDescending(s => s.HeuristicInfo).ToList<SelectedOrder>();
                    break;
                case 2:
                    for (int inser = 0; inser < BankList.Count; inser++)
                        BankList[inser].HeuristicInfo = BankList[inser].Order.TotalSetup * (1 + r.NextDouble());
                    BankList = BankList.OrderBy(s => s.HeuristicInfo).ToList<SelectedOrder>();
                    break;
                case 3:
                    for (int inser = 0; inser < BankList.Count; inser++)
                        BankList[inser].HeuristicInfo = BankList[inser].Order.FatherOrder.VTWList.Count * (1 + r.NextDouble());
                    BankList = BankList.OrderBy(s => s.HeuristicInfo).ToList<SelectedOrder>();
                    break;
                case 4:
                    for (int inser = 0; inser < BankList.Count; inser++)
                        BankList[inser].HeuristicInfo = BankList[inser].Order.Conflict * (1 + r.NextDouble());
                    BankList = BankList.OrderBy(s => s.HeuristicInfo).ToList<SelectedOrder>();
                    break;
            }

            SelectedOrder tempObs = new SelectedOrder();

            for (int k = 0; k < BankList.Count; k++)
            {

                tempObs = BankList[k];
                BankList.Remove(tempObs);
                BankRevenue -= tempObs.Order.Revenue;

                if (Upperbound + tempObs.Order.Revenue + BankRevenue < C_fit) //The upperbound of the current insertion is lower than the current solution
                    return;


                int IniLength = NewCurrent.Count;
                SelectedOrder SelectedObs = new SelectedOrder();
                CloneOrder(tempObs, SelectedObs);
                double newupperbound = 0;
                double DeltaTrans = InsertObsIntoSol(ref NewCurrent, SelectedObs, CurrentGeneList, Upperbound, out newupperbound);  //Insert the order

                if (NewCurrent.Count > IniLength)
                {//insert succeeds
                    Upperbound = newupperbound;
                    SelectedObs.Order.IsInNewCurrent = true;
                    SelectedObs.Order.FatherOrder.IsProcessed = true;
                    SelectedObs.JustInsert = true;
                    tempObs.Order.RemoveTabu = tabutime;
                }
                else
                {
                    UnScheduledOrders.Add(tempObs);
                }
                k = -1;
            }

            NC_fit = Problem.GetTotalProfit(NewCurrent);

            #region Solution update, operator score and weight update
            if (NC_fit > G_fit)//better than the best solution
            {
                IniSol.Solution = NewCurrent;
                IniSol.GeneList = CurrentGeneList;
                IniSol.Fitness = NC_fit;
                CurDelete.score += SigmaGlobal;
                CurInsert.score += SigmaGlobal;
                IniSol.UnScheduledOrders = UnScheduledOrders;
                UpdateGeneValue(IniSol);
            }
            else
            {
                if (NC_fit >= C_fit)//better than the current solution
                {
                    IniSol.Solution = NewCurrent;
                    IniSol.GeneList = CurrentGeneList;
                    IniSol.Fitness = NC_fit;
                    CurDelete.score += SigmaLocal;
                    CurInsert.score += SigmaLocal;
                    IniSol.UnScheduledOrders = UnScheduledOrders;
                    UpdateGeneValue(IniSol);
                }
                else//A worse solution might be accepted          
                {
                    double pp = Math.Exp(100 / Temperature * ((NC_fit - C_fit) / C_fit));
                    double temp = r.NextDouble();
                    if (pp >= temp)
                    {
                        IniSol.Solution = NewCurrent;
                        IniSol.GeneList = CurrentGeneList;
                        IniSol.Fitness = NC_fit;
                        CurDelete.score += SigmaBad;
                        CurInsert.score += SigmaBad;
                        IniSol.UnScheduledOrders = UnScheduledOrders;
                        UpdateGeneValue(IniSol);
                    }
                }
            }
            Temperature = Temperature * ParaOfAnnealing;//Annealing

            #endregion

            //Calculate good piece
            
            for (int i = 0; i < IniSol.Solution.Count - 2; i++)
            {
                double ShortPieceValue = (IniSol.Solution[i].Order.UnitRevenue + IniSol.Solution[i + 1].Order.UnitRevenue) / (IniSol.Solution[i + 1].RealStartTime - IniSol.Solution[i].RealEndTime);
                    
                if (ShortPieceValue > GoodPiece || (i == 0 && CurrentIte>=4 && ShortPieceValue>=firstgoodpiece))
                {
                    IniSol.GeneList[IniSol.Solution[i].Order.OrderID].IsPairedSuc = true;
                    IniSol.GeneList[IniSol.Solution[i + 1].Order.OrderID].IsPairedPre = true;
                    IniSol.GeneList[IniSol.Solution[i + 1].Order.OrderID].GeneValue = IniSol.GeneList[IniSol.Solution[i].Order.OrderID].GeneValue + 0.00001;
                    IniSol.GeneList[IniSol.Solution[i].Order.OrderID].PairGeneSuc = IniSol.GeneList[IniSol.Solution[i + 1].Order.OrderID];
                    IniSol.GeneList[IniSol.Solution[i + 1].Order.OrderID].PairGenePre = IniSol.GeneList[IniSol.Solution[i].Order.OrderID];
                }
            }
        }

        
        /// <summary>
        /// Update the gene value of all unscheduled orders
        /// </summary>
        /// <param name="OneChromo"></param>
        public void UpdateGeneValue(Chromosome OneChromo)
        {
            int count = 0;

            for (int i = 1; i < OneChromo.Solution.Count - 1; i++)
            {
                SelectedOrder a = OneChromo.Solution[i];
                OneChromo.GeneList[a.Order.OrderID].GeneValue = GetNewGeneValue(GeneValueLib[count],GeneValueLib[count+1]);
                count++;
            }
            for (int i = 0; i < OneChromo.UnScheduledOrders.Count; i++)
            {
                SelectedOrder a = OneChromo.UnScheduledOrders[i];
                OneChromo.GeneList[a.Order.OrderID].GeneValue = GetNewGeneValue(GeneValueLib[count], GeneValueLib[count + 1]);
                count++;
                double temp = Problem.GetEarliestTime(OneChromo.Solution[OneChromo.Solution.Count - 2], a);
                if (temp < a.Order.Deadline)
                {
                    a.RealStartTime = temp;
                    a.RealEndTime = temp + a.Order.ProcessTime;
                    a.Setup = Problem.GetSetupTime(OneChromo.Solution[OneChromo.Solution.Count - 2], a, 1, 1);
                    a.Profit = a.Order.Revenue - a.Order.Weight * Math.Max(a.RealEndTime - a.Order.DueTime, 0);
                    OneChromo.Solution.Insert(OneChromo.Solution.Count - 1, a);
                    OneChromo.Fitness += a.Profit;
                    OneChromo.UnScheduledOrders.Remove(a);
                    i--;
                }
            }
        }

        public void CloneGeneList(List<Gene> A, List<Gene> B)
        {
            B.Clear();
            foreach (Gene a in A)
            {
                Gene temp = new Gene();
                temp.GeneValue = a.GeneValue;
                temp.OrderId = a.OrderId;
                temp.PairGenePre = null;
                temp.PairGeneSuc = null;
                temp.IsPairedPre = false;
                temp.IsPairedSuc = false;
                temp.IsPairChosen = false;
                B.Add(temp);
            }
        }

        public void CloneSol(List<SelectedOrder> A, List<SelectedOrder> B)
        {
            B.Clear();
            foreach (SelectedOrder a in A)
            {
                SelectedOrder temp = new SelectedOrder();
                temp.DueTS = a.DueTS;
                temp.Order = a.Order;
                temp.Profit = a.Profit;
                temp.RealEndTime = a.RealEndTime;
                temp.RealStartTime = a.RealStartTime;
                temp.Setup = a.Setup;
                temp.TS = a.TS;
                B.Add(temp);
            }
        }

        /// <summary>
        /// Return all the possible insertion positions of the candidate order
        /// </summary>
        /// <param name="OldSolution">Current solution</param>
        /// <param name="ObsInsert">Candidate order</param>
        /// <param name="Time">The start time</param>
        /// <returns>All possible positions</returns>
        public List<position> GetPositionOfInsert(List<SelectedOrder> OldSolution, SelectedOrder ObsInsert, double Time)
        {
            List<position> positionlist = new List<position>();
            double TimePoint = Time;
            double TimePointEnd = 0;
            double folTimePoint = 0;
            double smalltrans = double.MaxValue;
            //List<position> pl = new List<position>();

            for (int i = 1; i < OldSolution.Count; i++)
            {
                if (OldSolution[i].Order.Deadline > Time)
                {
                    position tempposi = new position();
                    tempposi.posi = i;
                    TimePoint = Problem.GetEarliestTime(OldSolution[i - 1], ObsInsert);
                    TimePointEnd = TimePoint + ObsInsert.Order.ProcessTime;
                    folTimePoint = Problem.GetTempEarliestTime(TimePointEnd, ObsInsert, OldSolution[i]);

                    if (TimePoint > ObsInsert.Order.Deadline || folTimePoint - OldSolution[i].RealStartTime > OldSolution[i].TS)//Time slack is not enough
                        continue;

                    if (i == 1)
                        tempposi.trans = folTimePoint - TimePointEnd + TimePoint - OldSolution[i - 1].RealEndTime - (OldSolution[i].RealStartTime - OldSolution[i - 1].RealEndTime);
                    else
                        tempposi.trans = (Problem.GetSetupTime(ObsInsert, OldSolution[i], (ObsInsert.Order.Release + ObsInsert.Order.Deadline) / 2, (OldSolution[i].Order.Release + OldSolution[i].Order.Deadline) / 2)
                         + Problem.GetSetupTime(OldSolution[i - 1], ObsInsert, OldSolution[i - 1].RealEndTime, (ObsInsert.Order.Release + ObsInsert.Order.Deadline) / 2)
                         - Problem.GetSetupTime(OldSolution[i - 1], OldSolution[i], OldSolution[i - 1].RealEndTime, OldSolution[i].RealStartTime));

                    tempposi.heuristic = tempposi.trans;

                    if (Problem.IsDueTime)
                        if ((folTimePoint - OldSolution[i].RealStartTime > OldSolution[i].DueTS && i != OldSolution.Count - 1) || TimePoint + ObsInsert.Order.ProcessTime > ObsInsert.Order.DueTime)
                        {
                            tempposi.heuristic += 10000;
                        }

                    if (tempposi.heuristic < 1000 && tempposi.heuristic < smalltrans)
                    {
                        smalltrans = tempposi.heuristic;
                        positionlist.Clear();
                        positionlist.Add(tempposi);
                    }
                    if (tempposi.heuristic > 1000 && smalltrans > 1000)
                        positionlist.Add(tempposi);
                }
            }

            return positionlist;
        }

        /// <summary>
        /// Insert the candidate order to the current solution
        /// </summary>
        /// <param name="OldSolution">current solution</param>
        /// <param name="ObsInsert">candidate order</param>
        /// <param name="CurrentGeneList"></param>
        /// <param name="CurRev"></param>
        /// <param name="NewRev"></param>
        /// <returns></returns>
        public virtual double InsertObsIntoSol(ref List<SelectedOrder> OldSolution, SelectedOrder ObsInsert, List<Gene> CurrentGeneList, double CurRev, out double NewRev)
        {
            double DeltaTrans = 0;

            NewRev = CurRev;

            double TimePoint = ObsInsert.Order.Release;
            List<position> positionlist = GetPositionOfInsert(OldSolution, ObsInsert, TimePoint);
            int k = -100;

            if (positionlist.Count == 0)
                return 0;

            double OldStart = 0;
            double OldSetup = 0;

            if (positionlist[0].heuristic < 100) //The insertion will not reduce the profit of other orders
            {
                k = positionlist[0].posi;
                OldSetup = OldSolution[k].Setup;
                OldSolution.Insert(k, ObsInsert);

                for (int j = k; j < OldSolution.Count; j++)
                {
                    OldStart = OldSolution[j].RealStartTime;
                    OldSolution[j].RealStartTime = Problem.GetEarliestTime(OldSolution[j - 1], OldSolution[j]);
                    OldSolution[j].Setup = Problem.GetSetupTime(OldSolution[j - 1], OldSolution[j], OldSolution[j - 1].RealEndTime, OldSolution[j].RealStartTime);
                    if (OldSolution[j].RealStartTime == OldStart && j != k)
                        break;
                    OldSolution[j].TS -= (OldSolution[j].RealStartTime - OldStart);
                    OldSolution[j].DueTS -= (OldSolution[j].RealStartTime - OldStart);
                    OldSolution[j].RealEndTime = OldSolution[j].RealStartTime + OldSolution[j].Order.ProcessTime;
                }

                int flag = k;
                double SucStart = OldSolution[flag + 1].RealStartTime + OldSolution[flag + 1].TS;
                double SucDueStart = OldSolution[flag + 1].RealStartTime + OldSolution[flag + 1].DueTS;
                while (flag > 0)
                {
                    SucStart = Problem.GetLatestTime(SucStart, OldSolution[flag], OldSolution[flag + 1]);
                    double oldts = OldSolution[flag].TS;
                    double olddts = OldSolution[flag].DueTS;
                    OldSolution[flag].TS = Math.Min(OldSolution[flag].Order.Deadline - OldSolution[flag].RealStartTime,
                        SucStart - OldSolution[flag].RealStartTime);
                    if (Problem.IsDueTime)
                    {
                        SucDueStart = Problem.GetDueLatestTime(SucDueStart, OldSolution[flag], OldSolution[flag + 1]);
                        OldSolution[flag].DueTS = Math.Min(OldSolution[flag].Order.DueTime - OldSolution[flag].RealEndTime,
                            SucDueStart - OldSolution[flag].RealStartTime);
                    }
                    if (oldts == OldSolution[flag].TS && flag != k && olddts == OldSolution[flag].DueTS)
                    {
                        break;
                    }
                    flag--;
                }

                DeltaTrans = OldSolution[k].Setup + OldSolution[k + 1].Setup - OldSetup;
                CurrentGeneList[ObsInsert.Order.OrderID].GeneValue = GetNewGeneValue(CurrentGeneList[OldSolution[k - 1].Order.OrderID].GeneValue, CurrentGeneList[OldSolution[k + 1].Order.OrderID].GeneValue);
                NewRev = Problem.GetTotalProfit(OldSolution);
            }
            else //The insertion will reduce the profit of other orders because of the due time
            {
                double BestRev = CurRev;
                double BestDeltaTrans = 0;
                double TempGene = 0;
                List<SelectedOrder> BestTempSol = new List<SelectedOrder>();
                foreach (position a in positionlist)
                {
                    k = a.posi;
                    OldSetup = OldSolution[k].Setup;

                    List<SelectedOrder> TempSol = new List<SelectedOrder>();
                    CloneSol(OldSolution, TempSol);
                    TempSol.Insert(k, ObsInsert);
                    for (int j = k; j < TempSol.Count; j++)
                    {
                        OldStart = TempSol[j].RealStartTime;
                        TempSol[j].RealStartTime = Problem.GetEarliestTime(TempSol[j - 1], TempSol[j]);
                        TempSol[j].Setup = Problem.GetSetupTime(TempSol[j - 1], TempSol[j], TempSol[j - 1].RealEndTime, TempSol[j].RealStartTime);
                        if (TempSol[j].RealStartTime == OldStart && j != k)
                            break;
                        TempSol[j].RealEndTime = TempSol[j].RealStartTime + TempSol[j].Order.ProcessTime;
                    }
                    double TempFit = Problem.GetTotalProfit(TempSol);
                    //Problem.CheckSolution(TempSol);
                    if (TempFit > BestRev)
                    {
                        CloneSol(TempSol, BestTempSol);
                        BestRev = TempFit;
                        BestDeltaTrans = TempSol[k].Setup + TempSol[k + 1].Setup - OldSetup;
                        TempGene = GetNewGeneValue(CurrentGeneList[TempSol[k - 1].Order.OrderID].GeneValue, CurrentGeneList[TempSol[k + 1].Order.OrderID].GeneValue);
                    }
                    //Problem.CheckSolution(BestTempSol);
                }
                if (BestRev > CurRev)
                {
                    OldSolution = BestTempSol;
                    NewRev = BestRev;
                    CurrentGeneList[ObsInsert.Order.OrderID].GeneValue = TempGene;
                    AddTimeSlack(OldSolution);
                    DeltaTrans = BestDeltaTrans;
                }
            }
            return DeltaTrans;
        }

        /// <summary>
        /// Calculate the time slack of each job
        /// </summary>
        /// <param name="ListWithoutTS"></param>
        /// <returns></returns>
        public virtual List<SelectedOrder> AddTimeSlack(List<SelectedOrder> ListWithoutTS)
        {
            int i = ListWithoutTS.Count - 1;
            ListWithoutTS[i].TS = ListWithoutTS[i].Order.Deadline - ListWithoutTS[i].RealStartTime;
            if (Problem.IsDueTime)
                ListWithoutTS[i].DueTS = ListWithoutTS[i].Order.DueTime - ListWithoutTS[i].RealEndTime;
            double SucStart = ListWithoutTS[i].Order.Deadline;
            double SucDueStart = ListWithoutTS[i].Order.DueTime - ListWithoutTS[i].Order.ProcessTime;
            i--;
            while (i >= 0)
            {
                SucStart = Problem.GetLatestTime(SucStart, ListWithoutTS[i], ListWithoutTS[i + 1]);
                ListWithoutTS[i].TS = Math.Min(ListWithoutTS[i].Order.Deadline - ListWithoutTS[i].RealStartTime,
                    SucStart - ListWithoutTS[i].RealStartTime);
                if (Problem.IsDueTime)
                {
                    SucDueStart = Problem.GetDueLatestTime(SucDueStart, ListWithoutTS[i], ListWithoutTS[i + 1]);
                    ListWithoutTS[i].DueTS = Math.Min(ListWithoutTS[i].Order.DueTime - ListWithoutTS[i].RealEndTime,
                        SucDueStart - ListWithoutTS[i].RealStartTime);
                }
                i--;
            }
            return ListWithoutTS;
        }

        /// <summary>
        /// Update weights of operators
        /// </summary>
        /// <param name="NameOfList"></param>
        /// <param name="Lam"></param>
        public void UpdateHeuristicList(List<Heuristic> NameOfList, double Lam)
        {
            double total = 0;
            int M = NameOfList.Count;
            foreach (Heuristic a in NameOfList)
            {
                if (a.CalledTimes == 0)
                {
                    a.score = 10;
                    a.CalledTimes = 1;
                }
                a.score /= a.CalledTimes;
                total += a.score;
                a.CalledTimes = 0;
            }
            for (int i = 0; i < NameOfList.Count; i++)
            {
                double temp = NameOfList[i].score / total;
                NameOfList[i].weight = NameOfList[i].weight * Lam + (1 - Lam) * temp;
                NameOfList[i].score = 1;
            }
        }

        /// <summary>
        /// Update Solution job time and time slack
        /// </summary>
        /// <param name="NewObsList"></param>
        /// <param name="Banklist"></param>
        public virtual void UpdateSolution(ref List<SelectedOrder> NewObsList, List<SelectedOrder> Banklist)
        {
            for (int i = 1; i < NewObsList.Count - 1; i++)
            {
                NewObsList[i].RealStartTime = Problem.GetEarliestTime(NewObsList[i - 1], NewObsList[i]);
                NewObsList[i].RealEndTime = NewObsList[i].RealStartTime + NewObsList[i].Order.ProcessTime;
                NewObsList[i].Setup = Problem.GetSetupTime(NewObsList[i - 1], NewObsList[i], NewObsList[i - 1].RealEndTime, NewObsList[i].RealStartTime);
                if (NewObsList[i].RealStartTime > NewObsList[i].Order.Deadline)
                {
                    Banklist.Add(NewObsList[i]);
                    NewObsList.Remove(NewObsList[i]);
                    i--;
                    continue;
                }
            }
            AddTimeSlack(NewObsList);//Calculate time slacks
        }

        /// <summary>
        /// Select operators
        /// </summary>
        /// <param name="HeurList"></param>
        /// <returns></returns>
        public int SelectOfHeuristic(List<Heuristic> HeurList)
        {
            int HeuId = 0;
            double CurValue = r.NextDouble();
            double SumWeight = 0;
            for (int i = 0; i < HeurList.Count; i++)
            {
                SumWeight = SumWeight + HeurList[i].weight;
                if (CurValue <= SumWeight)
                {
                    HeuId = i;
                    break;
                }
            }
            return HeuId;
        }

        public virtual void CloneOrder(SelectedOrder a, SelectedOrder b)
        {
            b.TS = a.TS;
            b.DueTS = a.DueTS;
            b.Order = a.Order;
            b.Setup = a.Setup;
            b.Profit = a.Profit;
            b.RealEndTime = a.RealEndTime;
            b.RealStartTime = a.RealStartTime;
            b.JustInsert = false;
        }

        /// <summary>
        /// remove jobs
        /// </summary>
        /// <param name="cursol"></param>
        /// <param name="newsol"></param>
        /// <param name="Banklist"></param>
        /// <param name="CurrentGeneList"></param>
        /// <param name="JustRemove"></param>
        public void CheckRemove(List<SelectedOrder> cursol, List<SelectedOrder> newsol, List<SelectedOrder> Banklist, List<Gene> CurrentGeneList, List<Order> JustRemove)
        {
            newsol.Clear();
            for (int i = 0; i < cursol.Count; i++)
            {
                if (!cursol[i].ToBeRemoved)
                {
                    SelectedOrder temp = new SelectedOrder();
                    CloneOrder(cursol[i], temp);
                    newsol.Add(temp);
                }
                else
                {
                    Banklist.Add(cursol[i]);
                    JustRemove.Add(cursol[i].Order);
                    cursol[i].ToBeRemoved = false;
                    CurrentGeneList[cursol[i].Order.OrderID].GeneValue = GetNewGeneValue(CurrentGeneList[cursol[cursol.Count - 2].Order.OrderID].GeneValue, 1);
                }
            }
        }


        double GetNewGeneValue(double low, double high)
        {
            return (high - (high - low) * r.NextDouble());
        }

        public void CopyListToList(List<SelectedOrder> a, List<SelectedOrder> b)
        {
            if (b.Count != 0)
                b.Clear();
            foreach (SelectedOrder o in a)
                b.Add(o);
        }

        /// <summary>
        /// Initialize operators
        /// </summary>
        public void InitialOperatorPool()
        {
            for (int i = 0; i < Problem.DestroyList.Length; i++)
            {
                Heuristic a = new Heuristic();
                a.Heur_Id = Problem.DestroyList[i];
                decimal temp = (decimal)1 / Problem.DestroyList.Length; ;
                a.weight = (double)temp;
                a.type = "Destroy";
                a.score = 1;
                a.CalledTimes = 0;
                DestroyList.Add(a);
            }
            for (int i = 0; i < Problem.RepairList.Length; i++)
            {
                Heuristic a = new Heuristic();
                a.Heur_Id = Problem.RepairList[i];
                decimal temp = (decimal)1 / Problem.RepairList.Length; ;
                a.weight = (double)temp;
                a.type = "Repair";
                a.score = 1;
                a.CalledTimes = 0;
                InsertList.Add(a);
            }
        }
    }

    public class position
    {
        public int posi;
        public double trans;
        public double rand;
        public double heuristic;
    }
}

